#include "config.h"
#include "IniFile.h"

Config config;

void Config::Load()
{
    IniFile ini;
    std::string last_directory_open;
    std::string last_directory_save;
    std::string last_directory_export;
    std::string last_directory_import;
    std::vector<int> last_silence_select;

    ini.AddIntegerConstant("auto", CIPH_TYPE_AUTO);
    // Ignore error
    ini.LoadFromFile("config.ini", false);

    // General
    ini.GetStringValue("General", "last_directory_open", last_directory_open, "");
    ini.GetStringValue("General", "last_directory_save", last_directory_save, "");
    ini.GetStringValue("General", "last_directory_export", last_directory_export, "");
    ini.GetStringValue("General", "last_directory_import", last_directory_import, "");
    ini.GetIntegerValue("General", "last_silence_total", &this->last_silence_total);
    ini.GetMultipleIntegersValues("General", "last_silence_select", last_silence_select);

    for (int i : last_silence_select)
    {
        if (i < 0 || i >= this->last_silence_total)
        {
            last_silence_select.clear();
            break;
        }
    }

    this->last_directory_open = Utils::IsEmptyString(last_directory_open) ? QString() : Utils::StdStringToQString(last_directory_open);
    this->last_directory_save = Utils::IsEmptyString(last_directory_save) ? QString() : Utils::StdStringToQString(last_directory_save);
    this->last_directory_export = Utils::IsEmptyString(last_directory_export) ? QString() : Utils::StdStringToQString(last_directory_export);
    this->last_directory_import = Utils::IsEmptyString(last_directory_import) ? QString() : Utils::StdStringToQString(last_directory_import);
    this->last_silence_select = QVector<int>::fromStdVector(last_silence_select);

    // Decoder
    ini.GetBooleanValue("Decoder", "preserve_loop", &this->preserve_loop, true);
    ini.GetBooleanValue("Decoder", "use_float", &this->use_float, false);
    ini.GetIntegerValue("Decoder", "bit_depth", &this->bit_depth, 16);

    if (this->bit_depth != 8 && this->bit_depth != 16 && this->bit_depth != 24 && this->bit_depth != 32)
        this->bit_depth = 16;

    // Encoder
    ini.GetIntegerValue("Encoder", "quality", &this->quality, 0);
    ini.GetIntegerValue("Encoder", "cutoff_freq", &this->cutoff_freq, 0);

    if (this->quality < 0 || this->quality > 4)
        this->quality = 0;

    // import
    ini.GetIntegerValue("Import", "ciph_type", &this->ciph_type, CIPH_TYPE_AUTO);
    ini.GetBooleanValue("Import", "last_was_looped", &this->last_was_looped, false);
    ini.GetBooleanValue("Import", "use_input_loop", &this->use_input_loop, true);

    if (this->ciph_type != CIPH_TYPE_AUTO && this->ciph_type != 0 && this->ciph_type != 1)
        this->ciph_type = CIPH_TYPE_AUTO;
}

void Config::Save()
{
    IniFile ini;

    ini.AddIntegerConstant("auto", CIPH_TYPE_AUTO);
    ini.LoadFromFile("config.ini", false);

    // General
    ini.SetStringValue("General", "last_directory_open", Utils::QStringToStdString(this->last_directory_open));
    ini.SetStringValue("General", "last_directory_save", Utils::QStringToStdString(this->last_directory_save));
    ini.SetStringValue("General", "last_directory_export", Utils::QStringToStdString(this->last_directory_export));
    ini.SetStringValue("General", "last_directory_import", Utils::QStringToStdString(this->last_directory_import));
    ini.SetIntegerValue("General", "last_silence_total", this->last_silence_total);
    ini.SetMultipleIntegersValues("General", "last_silence_select", this->last_silence_select.toStdVector());

    // Decoder
    ini.SetBooleanValue("Decoder", "preserve_loop", this->preserve_loop);
    ini.SetBooleanValue("Decoder", "use_float", this->use_float);
    ini.SetIntegerValue("Decoder", "bit_depth", this->bit_depth);

    // Encoder
    ini.SetIntegerValue("Encoder", "quality", this->quality);
    ini.SetIntegerValue("Encoder", "cutoff_freq", this->cutoff_freq);

    // import
    ini.SetIntegerValue("Import", "ciph_type", this->ciph_type);
    ini.SetBooleanValue("Import", "last_was_looped", this->last_was_looped);
    ini.SetBooleanValue("Import", "use_input_loop", this->use_input_loop);

    ini.SaveToFile("config.ini");
}
