#include "optionsdialog.h"
#include "ui_optionsdialog.h"
#include "config.h"

OptionsDialog::OptionsDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::OptionsDialog)
{
    ui->setupUi(this);
    setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);

    ui->cutoffLineEdit->setValidator(new QIntValidator);

    ui->saveLoopCheckBox->setChecked(config.preserve_loop);
    if (config.use_float)
    {
        ui->formatComboBox->setCurrentIndex(4);
    }
    else if (config.bit_depth == 32)
    {
        ui->formatComboBox->setCurrentIndex(3);
    }
    else if (config.bit_depth == 24)
    {
        ui->formatComboBox->setCurrentIndex(2);
    }
    else if (config.bit_depth == 16)
    {
        ui->formatComboBox->setCurrentIndex(1);
    }
    else
    {
        ui->formatComboBox->setCurrentIndex(0);
    }

    ui->qualityComboBox->setCurrentIndex(config.quality);
    ui->cutoffLineEdit->setText(QString("%1").arg(config.cutoff_freq));

    ui->importLoopCheckBox->setChecked(config.use_input_loop);

    if (config.ciph_type == CIPH_TYPE_AUTO)
    {
        ui->ciphTypeComboBox->setCurrentIndex(0);
    }
    else if (config.ciph_type == 0)
    {
        ui->ciphTypeComboBox->setCurrentIndex(1);
    }
    else
    {
        ui->ciphTypeComboBox->setCurrentIndex(2);
    }
}

OptionsDialog::~OptionsDialog()
{
    delete ui;
}

void OptionsDialog::on_buttonBox_accepted()
{
    config.preserve_loop = ui->saveLoopCheckBox->isChecked();
    config.use_float = (ui->formatComboBox->currentIndex() == 4);

    if (ui->formatComboBox->currentIndex() == 3)
    {
        config.bit_depth = 32;
    }
    else if (ui->formatComboBox->currentIndex() == 2)
    {
        config.bit_depth = 24;
    }
    else if (ui->formatComboBox->currentIndex() == 1)
    {
        config.bit_depth = 16;
    }
    else if (ui->formatComboBox->currentIndex() == 0)
    {
        config.bit_depth = 8;
    }

    config.quality = ui->qualityComboBox->currentIndex();
    config.cutoff_freq = ui->cutoffLineEdit->text().toLong();
    config.use_input_loop = ui->importLoopCheckBox->isChecked();

    if (ui->ciphTypeComboBox->currentIndex() == 0)
    {
        config.ciph_type = CIPH_TYPE_AUTO;
    }
    else if (ui->ciphTypeComboBox->currentIndex() == 1)
    {
        config.ciph_type = 0;
    }
    else
    {
        config.ciph_type = 1;
    }

    done(1);
}

void OptionsDialog::on_buttonBox_rejected()
{
    done(0);
}
