#include "IdbFile.h"
#include "debug.h"

IdbFile::IdbFile()
{

}

IdbFile::~IdbFile()
{

}

void IdbFile::Reset()
{
    entries.clear();
}

bool IdbFile::Load(const uint8_t *buf, size_t size)
{
    Reset();

    if (!buf || size < sizeof(IDBHeader))
        return false;

    const IDBHeader *hdr = (const IDBHeader *)buf;

    if (hdr->signature != IDB_SIGNATURE)
        return false;

    const IDBEntry *file_entries = (const IDBEntry *)(buf + hdr->data_start);
    entries.resize(hdr->num_entries);

    for (size_t i = 0; i < entries.size(); i++)
    {
        IdbEntry &entry = entries[i];

        entry.id = file_entries[i].id;
        entry.stars = file_entries[i].stars;
        entry.name_id = file_entries[i].name_id;
        entry.desc_id = file_entries[i].desc_id;
        entry.type = file_entries[i].type;
        entry.unk_0A = file_entries[i].unk_0A;
        entry.unk_0C = file_entries[i].unk_0C;
        entry.unk_0E = file_entries[i].unk_0E;
        entry.buy = file_entries[i].buy;
        entry.sell = file_entries[i].sell;
        entry.racelock = file_entries[i].racelock;
        entry.tp = file_entries[i].tp;
        entry.model = file_entries[i].model;
        memcpy(entry.unk_24, file_entries[i].unk_24, sizeof(entry.unk_24));
        memcpy(entry.set1, file_entries[i].set1, sizeof(entry.set1));
        memcpy(entry.set2, file_entries[i].set2, sizeof(entry.set2));
        memcpy(entry.set3, file_entries[i].set3, sizeof(entry.set3));
    }

    return true;
}

IdbEntry *IdbFile::FindEntryByID(uint32_t id)
{
    for (IdbEntry &entry : entries)
    {
        if (entry.id == id)
            return &entry;
    }

    return nullptr;
}
