#ifndef __IDBFILE_H__
#define __IDBFILE_H__

#include "BaseFile.h"

#define IDB_SIGNATURE   0x42444923

#ifdef _MSC_VER
#pragma pack(push,1)
#endif

typedef struct
{
    uint32_t signature; // 0
    uint16_t endianess_check; // 4
    uint16_t unk_06;
    uint32_t num_entries; // 8
    uint32_t data_start; // 0xC
} PACKED IDBHeader;

STATIC_ASSERT_STRUCT(IDBHeader, 0x10);

typedef struct
{
    uint16_t id; // 0
    uint16_t stars; // 2
    uint16_t name_id; // 4
    uint16_t desc_id; // 6
    uint16_t type; // 8
    uint16_t unk_0A;
    uint16_t unk_0C;
    uint16_t unk_0E;
    uint32_t buy; // 0x10
    uint32_t sell; // 0x14
    uint32_t racelock; // 0x18
    uint32_t tp; // 0x1C
    uint32_t model; // 0x20
    uint32_t unk_24[3]; //
    uint8_t set1[0xE0]; // 0x30
    uint8_t set2[0xE0]; // 0x110
    uint8_t set3[0xE0]; // 0x1F0
} PACKED IDBEntry;

STATIC_ASSERT_STRUCT(IDBEntry, 0x2D0);

#ifdef _MSC_VER
#pragma pack(pop)
#endif


struct IdbEntry
{
    uint16_t id;
    uint16_t stars;
    uint16_t name_id;
    uint16_t desc_id;
    uint16_t type;
    uint16_t unk_0A;
    uint16_t unk_0C;
    uint16_t unk_0E;
    uint32_t buy;
    uint32_t sell;
    uint32_t racelock;
    uint32_t tp;
    uint32_t model;
    uint32_t unk_24[3];
    uint8_t set1[0xE0];
    uint8_t set2[0xE0];
    uint8_t set3[0xE0];
};

class IdbFile : public BaseFile
{
private:

    std::vector<IdbEntry> entries;

    void Reset();

public:

    IdbFile();
    virtual ~IdbFile();

    virtual bool Load(const uint8_t *buf, size_t size) override;
    //virtual uint8_t *Save(size_t *psize) override;

    //virtual TiXmlDocument *Decompile() const override;
    //virtual bool Compile(TiXmlDocument *doc, bool big_endian=false) override;

    inline size_t GetNumEntries() const { return entries.size(); }

    IdbEntry *FindEntryByID(uint32_t id);

    inline const std::vector<IdbEntry> &GetEntries() const { return entries; }
    inline std::vector<IdbEntry> &GetEntries() { return entries; }

    inline const IdbEntry &operator[](size_t n) const { return entries[n]; }
    inline const IdbEntry &operator[](size_t n) { return entries[n]; }

    inline std::vector<IdbEntry>::const_iterator begin() const { return entries.begin(); }
    inline std::vector<IdbEntry>::const_iterator end() const { return entries.end(); }

    inline std::vector<IdbEntry>::iterator begin() { return entries.begin(); }
    inline std::vector<IdbEntry>::iterator end() { return entries.end(); }
};

#endif // IDBFILE_H
