#include <algorithm>

#include "Xenoverse2.h"
#include "SwfFile.h"
#include "debug.h"

#undef ReplaceFile

#define CHARALIST_PATH  "/action_script/CharaList.as"
#define TIMELINE_PATH   "/CHARASELE_fla/MainTimeline.as"
#define COMPRESULT_PATH "/CHARASELE_fla/MainTimeline.swf"

#define GAME_CHARASELE_PATH     "data/ui/iggy/CHARASELE.iggy"
#define GAME_SELPORT_PATH       "data/ui/texture/CHARA01.emb"
#define GAME_CHARANAMES_PATH    "data/msg/proper_noun_character_name_"
#define GAME_CHARA_COSTUMENAMES_PATH  "data/msg/proper_noun_variation_name_"

#define GAME_CMS_PATH   "data/system/char_model_spec.cms"
#define GAME_CUS_PATH   "data/system/custom_skill.cus"
#define GAME_CSO_PATH   "data/system/chara_sound.cso"
#define GAME_PSC_PATH   "data/system/parameter_spec_char.psc"
#define GAME_AUR_PATH   "data/system/aura_setting.aur"
#define GAME_SEV_PATH   "data/system/special_event_voice.sev"
#define GAME_CML_PATH   "data/system/CameraLimitValue.cml"
#define GAME_HCI_PATH   "data/ui/CharaImage/chara_image.hci"
#define GAME_PREBAKED_PATH  "data/pre-baked.xml"

#define GAME_LOBBY_TEXT_PATH    "data/msg/lobby_text_"
#define GAME_CAC_COSTUMENAMES_PATH  "data/msg/proper_noun_costume_name_"
#define GAME_ACCESORY_NAMES_PATH    "data/msg/proper_noun_accessory_name_"
#define GAME_TALISMAN_NAMES_PATH    "data/msg/proper_noun_talisman_name_"

#define GAME_BOTTOM_IDB_PATH "data/system/item/costume_bottom_item.idb"
#define GAME_GLOVES_IDB_PATH "data/system/item/costume_gloves_item.idb"
#define GAME_SHOES_IDB_PATH "data/system/item/costume_shoes_item.idb"
#define GAME_TOP_IDB_PATH "data/system/item/costume_top_item.idb"
#define GAME_ACCESORY_IDB_PATH "data/system/item/accessory_item.idb"
#define GAME_TALISMAN_IDB_PATH "data/system/item/talisman_item.idb"

#define GAME_PAL_PATH   "data/event/lobby/lobby_preset_avatar_list.pal"

#define GAME_CSS_SOUND_JP_PATH  "data/sound/VOX/Sys/CRT_CS_vox"
#define GAME_CSS_SOUND_EN_PATH  "data/sound/VOX/Sys/en/CRT_CS_vox"

const std::vector<std::string> xv2_lang_codes =
{
    "en",
    "es",
    "ca",
    "fr",
    "de",
    "it",
    "pt",
    "pl",
    "ru"
};

static const std::vector<std::string> original_chars =
{
    "CMN", // Not a char, but we want it in this list. Keep it at index 0!
    "APL",
    "BAT",
    "BDK",
    "BLM",
    "BLS",
    "BRL",
    "BUL",
    "BUM",
    "BUS",
    "BUZ",
    "CAB",
    "CL1",
    "CL3",
    "CL4",
    "CLJ",
    "CO2",
    "COL",
    "CRC",
    "DDR",
    "DND",
    "FOF",
    "FR4",
    "FR5",
    "FRG",
    "FRI",
    "FRS",
    "FRZ",
    "G16",
    "G17",
    "G18",
    "GFR",
    "GG1",
    "GGK",
    "GGT",
    "GHF",
    "GHL",
    "GHM",
    "GHS",
    "GK4",
    "GKB",
    "GKG",
    "GNY",
    "GOD",
    "GOK",
    "GRD",
    "GTG",
    "GTX",
    "GVG",
    "HIK",
    "HIT",
    "HST",
    "HUF",
    "HUM",
    "JCO",
    "JES",
    "JNB",
    "JNG",
    "KLL",
    "MAF",
    "MAM",
    "MAP",
    "MIR",
    "MRN",
    "NAP",
    "NIL",
    "NMC",
    "NP1",
    "NP2",
    "NP3",
    "NP4",
    "OBB",
    "OSB",
    "OSN",
    "OSV",
    "PAN",
    "PIC",
    "POD",
    "RAD",
    "RCM",
    "ROK",
    "RSB",
    "S17",
    "SBM",
    "SCR",
    "SD1",
    "SD3",
    "SD4",
    "SIN",
    "SLG",
    "STD",
    "STN",
    "SYF",
    "SYM",
    "TG1",
    "TG2",
    "TG3",
    "TG4",
    "TG5",
    "TKT",
    "TLS",
    "TOK",
    "TOW",
    "TRF",
    "TRS",
    "TRX",
    "TSH",
    "VDL",
    "VG4",
    "VGT",
    "VTO",
    "WIS",
    "YMC",
    "ZBN"
};

Xv2Fs *xv2fs;
CharaListFile *chara_list;
IggyFile *charasele;

CmsFile *game_cms;
CusFile *game_cus;
CsoFile *game_cso;
PscFile *game_psc;
AurFile *game_aur;
SevFile *game_sev;
CmlFile *game_cml;
HciFile *game_hci;

std::vector<MsgFile *> game_chara_names;
std::vector<MsgFile *> game_chara_costume_names;

std::vector<MsgFile *> sup_sk_names;
std::vector<MsgFile *> ult_sk_names;
std::vector<MsgFile *> eva_sk_names;
std::vector<MsgFile *> awa_sk_names;

EmbFile *game_sel_port;
Xv2PreBakedFile *game_prebaked;

std::vector<MsgFile *> game_lobby_texts;
std::vector<MsgFile *> game_cac_costume_names;
std::vector<MsgFile *> game_accesory_names;
std::vector<MsgFile *> game_talisman_names;

IdbFile *game_bottom_idb;
IdbFile *game_gloves_idb;
IdbFile *game_shoes_idb;
IdbFile *game_top_idb;
IdbFile *game_accesory_idb;
IdbFile *game_talisman_idb;

PalFile *game_pal;

AcbFile *game_css_acb_jp;
AcbFile *game_css_acb_en;
Afs2File *game_css_awb_jp;
Afs2File *game_css_awb_en;

static std::string chasel_path, chalist_path;

struct HtmlEscape
{
    std::u16string sequence;
    char16_t c;
};

// These were all sequences found in all msg files of the game
static const std::vector<HtmlEscape> xv2_html_sequences =
{
    // HACK: VS2013 doesn't support u16 literals
    { (char16_t *)L"&quot;", '"' },
    { (char16_t *)L"&apos;", '\'' },
    { (char16_t *)L"&gt;", '>' },
    { (char16_t *)L"&lt;", '<' },
    { (char16_t *)L"&amp;", '&' },
};

static inline std::string UnescapeHtml(const std::string &str)
{
    if (str.find('&') == std::string::npos || str.find(';') == std::string::npos)
        return str;

    std::u16string str16 = Utils::Utf8ToUcs2(str);

    for (size_t i = 0; i < str16.length()-3; i++)
    {
        if (str16[i] == '&')
        {
            for (const HtmlEscape &esc : xv2_html_sequences)
            {
                if (str16.find(esc.sequence, i) == i)
                {
                    str16.replace(i, esc.sequence.length(), std::u16string { esc.c });
                }
            }
        }
    }

    return Utils::Ucs2ToUtf8(str16);
}

static inline std::string EscapeHtml(const std::string &str)
{
    bool found = false;

    for (const HtmlEscape &esc : xv2_html_sequences)
    {
        if (str.find((char)esc.c) != std::string::npos)
        {
            found = true;
            break;
        }
    }

    if (!found)
        return str;

    std::u16string str16 = Utils::Utf8ToUcs2(str);

    for (size_t i = 0; i < str16.length(); i++)
    {
        for (const HtmlEscape &esc : xv2_html_sequences)
        {
            if (str16[i] == esc.c)
            {
                str16.replace(i, 1, esc.sequence);
                i += esc.sequence.length()-1;
            }
        }
    }

    return Utils::Ucs2ToUtf8(str16);
}

void Xenoverse2::InitFs(const std::string &game_path)
{
    if (xv2fs)
        delete xv2fs;

    xv2fs = new Xv2Fs(game_path);
}

bool Xenoverse2::InitCharaList(const std::string &path, const std::string &list_path)
{
    chasel_path = path;
    chalist_path = list_path;

    if (chara_list)
        delete chara_list;

    chara_list = new CharaListFile();
    if (!chara_list->LoadFromFile(chalist_path))
    {
        delete chara_list;
        chara_list = nullptr;
        return false;
    }

    if (charasele)
        delete charasele;

    charasele = new IggyFile();

    if (!xv2fs->LoadFile(charasele, GAME_CHARASELE_PATH))
    {
        delete chara_list; chara_list = nullptr;
        delete charasele; charasele = nullptr;
        return false;
    }

    return true;
}

bool Xenoverse2::CommitCharaList()
{
    if (!xv2fs || !chara_list || !charasele)
        return false;

    if (!xv2fs->SaveFile(charasele, GAME_CHARASELE_PATH))
        return false;

    return chara_list->SaveToFile(chalist_path, true, true);
}

bool Xenoverse2::InitSystemFiles(bool only_cms)
{
    if (!xv2fs)
        return false;

    if (game_cms)
        delete game_cms;

    game_cms = new CmsFile();
    if (!xv2fs->LoadFile(game_cms, GAME_CMS_PATH) || !game_cms->IsXV2())
    {
        DPRINTF("%s: cannot load cms.\n", FUNCNAME);
        delete game_cms; game_cms = nullptr;
        return false;
    }

    if (only_cms)
        return true;

    if (game_cus)
        delete game_cus;

    game_cus = new CusFile();
    if (!xv2fs->LoadFile(game_cus, GAME_CUS_PATH))
    {
        DPRINTF("%s: cannot load cus.\n", FUNCNAME);
        delete game_cus; game_cus = nullptr;
        return false;
    }

    if (game_cso)
        delete game_cso;

    game_cso = new CsoFile();
    if (!xv2fs->LoadFile(game_cso, GAME_CSO_PATH))
    {
        DPRINTF("%s: cannot load cso.\n", FUNCNAME);
        delete game_cso; game_cso = nullptr;
        return false;
    }

    if (game_psc)
        delete game_psc;

    game_psc = new PscFile();
    if (!xv2fs->LoadFile(game_psc, GAME_PSC_PATH))
    {
        DPRINTF("%s: cannot load psc.\n", FUNCNAME);
        delete game_psc; game_psc = nullptr;
        return false;
    }

    if (game_aur)
        delete game_aur;

    game_aur = new AurFile();
    if (!xv2fs->LoadFile(game_aur, GAME_AUR_PATH))
    {
        DPRINTF("%s: cannot load aur.\n", FUNCNAME);
        delete game_aur; game_aur = nullptr;
        return false;
    }

    if (game_sev)
        delete game_sev;

    game_sev = new SevFile();
    if (!xv2fs->LoadFile(game_sev, GAME_SEV_PATH))
    {
        DPRINTF("%s: cannot load sev.\n", FUNCNAME);
        delete game_sev; game_sev = nullptr;
        return false;
    }

    if (game_cml)
        delete game_cml;

    game_cml = new CmlFile();
    if (!xv2fs->LoadFile(game_cml, GAME_CML_PATH))
    {
        DPRINTF("%s: cannot load cml.\n", FUNCNAME);
        delete game_cml; game_cml = nullptr;
        return false;
    }

    if (game_hci)
        delete game_hci;

    game_hci = new HciFile();
    if (!xv2fs->LoadFile(game_hci, GAME_HCI_PATH))
    {
        DPRINTF("%s: cannot load hci.\n", FUNCNAME);
        delete game_hci; game_hci = nullptr;
        return false;
    }

    return true;
}

bool Xenoverse2::CommitSystemFiles()
{
    if (!xv2fs || !game_cms || !game_cus || !game_cso || !game_psc || !game_aur || !game_sev || !game_cml || !game_hci)
        return false;

    if (!xv2fs->SaveFile(game_cms, GAME_CMS_PATH))
    {
        DPRINTF("Save of cms failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_cus, GAME_CUS_PATH))
    {
        DPRINTF("Save of cus failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_cso, GAME_CSO_PATH))
    {
        DPRINTF("Save of cso failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_psc, GAME_PSC_PATH))
    {
        DPRINTF("Save of psc failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_aur, GAME_AUR_PATH))
    {
        DPRINTF("Save of aur failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_sev, GAME_SEV_PATH))
    {
        DPRINTF("Save of sev failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_cml, GAME_CML_PATH))
    {
        DPRINTF("Save of cml failed.\n");
        return false;
    }

    if (!xv2fs->SaveFile(game_hci, GAME_HCI_PATH))
    {
        DPRINTF("Save of hci failed.\n");
        return false;
    }

    return true;
}

bool Xenoverse2::InitCharaNames(int only_this_lang)
{
    if (!xv2fs)
        return false;

    for (MsgFile *msg : game_chara_names)
    {
        if (msg)
            delete msg;
    }

    game_chara_names.clear();
    game_chara_names.resize(XV2_LANG_NUM);

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (only_this_lang >= 0 && i != only_this_lang)
            continue;

        std::string path = GAME_CHARANAMES_PATH + xv2_lang_codes[i] + ".msg";
        game_chara_names[i] = new MsgFile();

        if (!xv2fs->LoadFile(game_chara_names[i], path))
        {
            DPRINTF("Failed loading file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::CommitCharaNames()
{
    if (!xv2fs || game_chara_names.size() != XV2_LANG_NUM)
        return false;

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        if (!game_chara_names[i])
            continue;

        std::string path = GAME_CHARANAMES_PATH + xv2_lang_codes[i] + ".msg";

        if (!xv2fs->SaveFile(game_chara_names[i], path))
        {
            DPRINTF("Failed saving file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::InitCharaCostumeNames(int only_this_lang)
{
    if (!xv2fs)
        return false;

    for (MsgFile *msg : game_chara_costume_names)
    {
        if (msg)
            delete msg;
    }

    game_chara_costume_names.clear();
    game_chara_costume_names.resize(XV2_LANG_NUM);

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (only_this_lang >= 0 && i != only_this_lang)
            continue;

        std::string path = GAME_CHARA_COSTUMENAMES_PATH + xv2_lang_codes[i] + ".msg";
        game_chara_costume_names[i] = new MsgFile();

        if (!xv2fs->LoadFile(game_chara_costume_names[i], path))
        {
            DPRINTF("Failed loading file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::CommitCharaCostumeNames()
{
    if (!xv2fs || game_chara_costume_names.size() != XV2_LANG_NUM)
        return false;

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        if (!game_chara_costume_names[i])
            continue;

        std::string path = GAME_CHARA_COSTUMENAMES_PATH + xv2_lang_codes[i] + ".msg";

        if (!xv2fs->SaveFile(game_chara_costume_names[i], path))
        {
            DPRINTF("Failed saving file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::InitSkillNames(int only_this_lang)
{
    if (!xv2fs)
        return false;

    for (MsgFile *msg : sup_sk_names)
    {
        if (msg)
            delete msg;
    }

    for (MsgFile *msg : ult_sk_names)
    {
        if (msg)
            delete msg;
    }

    for (MsgFile *msg : eva_sk_names)
    {
        if (msg)
            delete msg;
    }

    for (MsgFile *msg : awa_sk_names)
    {
        if (msg)
            delete msg;
    }

    sup_sk_names.clear();
    ult_sk_names.clear();
    eva_sk_names.clear();
    awa_sk_names.clear();

    sup_sk_names.resize(XV2_LANG_NUM);
    ult_sk_names.resize(XV2_LANG_NUM);
    eva_sk_names.resize(XV2_LANG_NUM);
    awa_sk_names.resize(XV2_LANG_NUM);

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (only_this_lang >= 0 && i != only_this_lang)
            continue;

        {
            std::string path = "data/msg/proper_noun_skill_spa_name_" + xv2_lang_codes[i] + ".msg";
            sup_sk_names[i] = new MsgFile();

            if (!xv2fs->LoadFile(sup_sk_names[i], path))
                return false;
        }

        {
            std::string path = "data/msg/proper_noun_skill_ult_name_" + xv2_lang_codes[i] + ".msg";
            ult_sk_names[i] = new MsgFile();

            if (!xv2fs->LoadFile(ult_sk_names[i], path))
                return false;
        }

        {
            std::string path = "data/msg/proper_noun_skill_esc_name_" + xv2_lang_codes[i] + ".msg";
            eva_sk_names[i] = new MsgFile();

            if (!xv2fs->LoadFile(eva_sk_names[i], path))
                return false;
        }

        {
            std::string path = "data/msg/proper_noun_skill_met_name_" + xv2_lang_codes[i] + ".msg";
            awa_sk_names[i] = new MsgFile();

            if (!xv2fs->LoadFile(awa_sk_names[i], path))
                return false;
        }
    }

    return true;
}

bool Xenoverse2::InitSelPort()
{
    if (!xv2fs)
        return false;

    if (game_sel_port)
        delete game_sel_port;

    game_sel_port = new EmbFile();
    if (!xv2fs->LoadFile(game_sel_port, GAME_SELPORT_PATH))
    {
        delete game_sel_port;
        game_sel_port = nullptr;
        return false;
    }

    return true;
}

bool Xenoverse2::CommitSelPort()
{
    if (!xv2fs || !game_sel_port)
        return false;

    return xv2fs->SaveFile(game_sel_port, GAME_SELPORT_PATH);
}

bool Xenoverse2::InitPreBaked()
{
    if (!xv2fs)
        return false;

    if (game_prebaked)
        delete game_prebaked;

    game_prebaked = new Xv2PreBakedFile();
    xv2fs->CompileFile(game_prebaked, GAME_PREBAKED_PATH); // Ignore the error, as file may not exist

    return true;
}

bool Xenoverse2::CommitPreBaked()
{
    if (!xv2fs || !game_prebaked)
        return false;

    return xv2fs->DecompileFile(game_prebaked, GAME_PREBAKED_PATH);
}

bool Xenoverse2::InitLobbyText(int only_this_lang)
{
    if (!xv2fs)
        return false;

    for (MsgFile *msg : game_lobby_texts)
    {
        if (msg)
            delete msg;
    }

    game_lobby_texts.clear();
    game_lobby_texts.resize(XV2_LANG_NUM);

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (only_this_lang >= 0 && i != only_this_lang)
            continue;

        std::string path = GAME_LOBBY_TEXT_PATH + xv2_lang_codes[i] + ".msg";
        game_lobby_texts[i] = new MsgFile();

        if (!xv2fs->LoadFile(game_lobby_texts[i], path))
        {
            DPRINTF("Failed loading file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::CommitLobbyText()
{
    if (!xv2fs || game_lobby_texts.size() != XV2_LANG_NUM)
        return false;

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        if (!game_lobby_texts[i])
            continue;

        std::string path = GAME_LOBBY_TEXT_PATH + xv2_lang_codes[i] + ".msg";

        if (!xv2fs->SaveFile(game_lobby_texts[i], path))
        {
            DPRINTF("Failed saving file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::InitCacCostumeNames(int only_this_lang)
{
    if (!xv2fs)
        return false;

    for (MsgFile *msg : game_cac_costume_names)
    {
        if (msg)
            delete msg;
    }

    game_cac_costume_names.clear();
    game_accesory_names.clear();

    game_cac_costume_names.resize(XV2_LANG_NUM);
    game_accesory_names.resize(XV2_LANG_NUM);

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (only_this_lang >= 0 && i != only_this_lang)
            continue;

        {
            std::string path = GAME_CAC_COSTUMENAMES_PATH + xv2_lang_codes[i] + ".msg";
            game_cac_costume_names[i] = new MsgFile();

            if (!xv2fs->LoadFile(game_cac_costume_names[i], path))
                return false;
        }

        {
            std::string path = GAME_ACCESORY_NAMES_PATH + xv2_lang_codes[i] + ".msg";
            game_accesory_names[i] = new MsgFile();            

            if (!xv2fs->LoadFile(game_accesory_names[i], path))
                return false;            
        }
    }

    return true;
}

bool Xenoverse2::InitTalismanNames(int only_this_lang)
{
    if (!xv2fs)
        return false;

    for (MsgFile *msg : game_talisman_names)
    {
        if (msg)
            delete msg;
    }

    game_talisman_names.clear();
    game_talisman_names.resize(XV2_LANG_NUM);

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (only_this_lang >= 0 && i != only_this_lang)
            continue;

        std::string path = GAME_TALISMAN_NAMES_PATH + xv2_lang_codes[i] + ".msg";
        game_talisman_names[i] = new MsgFile();

        if (!xv2fs->LoadFile(game_talisman_names[i], path))
        {
            DPRINTF("Failed loading file \"%s\"\n", path.c_str());
            return false;
        }
    }

    return true;
}

bool Xenoverse2::InitIdb(bool costumes, bool accesories, bool talisman)
{
    if (!xv2fs)
        return false;

    if (costumes)
    {
        if (game_top_idb)
            delete game_top_idb;        

        game_top_idb = new IdbFile();
        if (!xv2fs->LoadFile(game_top_idb, GAME_TOP_IDB_PATH))
        {
            delete game_top_idb; game_top_idb = nullptr;
            return false;
        }

        if (game_bottom_idb)
            delete game_bottom_idb;        

        game_bottom_idb = new IdbFile();
        if (!xv2fs->LoadFile(game_bottom_idb, GAME_BOTTOM_IDB_PATH))
        {
            delete game_bottom_idb; game_bottom_idb = nullptr;
            return false;
        }

        if (game_gloves_idb)
            delete game_gloves_idb;

        game_gloves_idb = new IdbFile();
        if (!xv2fs->LoadFile(game_gloves_idb, GAME_GLOVES_IDB_PATH))
        {
            delete game_gloves_idb; game_gloves_idb = nullptr;
            return false;
        }        

        if (game_shoes_idb)
            delete game_shoes_idb;

        game_shoes_idb = new IdbFile();
        if (!xv2fs->LoadFile(game_shoes_idb, GAME_SHOES_IDB_PATH))
        {
            delete game_shoes_idb; game_shoes_idb = nullptr;
            return false;
        }
    }

    if (accesories)
    {
        if (game_accesory_idb)
            delete game_accesory_idb;

        game_accesory_idb = new IdbFile();

        if (!xv2fs->LoadFile(game_accesory_idb, GAME_ACCESORY_IDB_PATH))
        {
            delete game_accesory_idb; game_accesory_idb = nullptr;
            return false;
        }
    }

    if (talisman)
    {
        if (game_talisman_idb)
            delete game_talisman_idb;

        game_talisman_idb = new IdbFile();

        if (!xv2fs->LoadFile(game_talisman_idb, GAME_TALISMAN_IDB_PATH))
        {
            delete game_talisman_idb; game_talisman_idb = nullptr;
            return false;
        }
    }

    return true;
}

bool Xenoverse2::InitLobby()
{
    if (!xv2fs)
        return false;

    if (game_pal)
        delete game_pal;

    game_pal = new PalFile();
    if (!xv2fs->LoadFile(game_pal, GAME_PAL_PATH))
        return false;

    return true;
}

bool Xenoverse2::CommitLobby()
{
    if (!xv2fs || !game_pal)
        return false;

    if (!xv2fs->SaveFile(game_pal, GAME_PAL_PATH))
        return false;

    return true;
}

bool Xenoverse2::InitSound()
{
    if (!xv2fs)
        return false;

    if (game_css_acb_jp)
        delete game_css_acb_jp;

    game_css_acb_jp = new AcbFile();
    if (!xv2fs->LoadFile(dynamic_cast<CriwareAudioContainer *>(game_css_acb_jp), std::string(GAME_CSS_SOUND_JP_PATH) + ".acb"))
    {
        delete game_css_acb_jp; game_css_acb_jp = nullptr;
        return false;
    }

    if (game_css_acb_en)
        delete game_css_acb_en;

    game_css_acb_en = new AcbFile();
    if (!xv2fs->LoadFile(dynamic_cast<CriwareAudioContainer *>(game_css_acb_en), std::string(GAME_CSS_SOUND_EN_PATH) + ".acb"))
    {
        delete game_css_acb_en; game_css_acb_en = nullptr;
        return false;
    }

    if (game_css_awb_jp)
        delete game_css_awb_jp;

    game_css_awb_jp = new Afs2File();
    if (!xv2fs->LoadFile(game_css_awb_jp, std::string(GAME_CSS_SOUND_JP_PATH) + ".awb"))
    {
        delete game_css_awb_jp; game_css_awb_jp = nullptr;
        return false;
    }

    if (game_css_awb_en)
        delete game_css_awb_en;

    game_css_awb_en = new Afs2File();
    if (!xv2fs->LoadFile(game_css_awb_en, std::string(GAME_CSS_SOUND_EN_PATH) + ".awb"))
    {
        delete game_css_awb_en; game_css_awb_en = nullptr;
        return false;
    }

    return true;
}

static bool SetAcbAwbData(AcbFile *acb, AwbFile *awb)
{
    if (acb->HasAwbHeader())
    {
        unsigned int header_size;
        uint8_t *awb_header = awb->CreateHeader(&header_size);
        if (!awb_header)
        {
            DPRINTF("%s: Failed to create awb header.\n", FUNCNAME);
            return false;
        }

        if (!acb->SetAwbHeader(awb_header, header_size, true))
        {
            delete[] awb_header;
            DPRINTF("Failed to set awb header in .acb.\n");
            return false;
        }
    }

    uint8_t md5[16];
    size_t size;
    uint8_t *buf = awb->Save(&size);

    if (!buf)
    {
        DPRINTF("%s: cannot get awb md5 because save file failed.\n", FUNCNAME);
        return false;
    }

    Utils::Md5(buf, (uint32_t)size, md5);
    acb->SetExternalAwbHash(md5);

    return true;
}

bool Xenoverse2::CommitSound()
{
    if (!xv2fs || !game_css_acb_jp || !game_css_acb_en || !game_css_awb_jp || !game_css_awb_en)
        return false;

    if (!SetAcbAwbData(game_css_acb_jp, game_css_awb_jp))
        return false;

    if (!SetAcbAwbData(game_css_acb_en, game_css_awb_en))
        return false;

    if (!xv2fs->SaveFile(dynamic_cast<CriwareAudioContainer *>(game_css_acb_jp), std::string(GAME_CSS_SOUND_JP_PATH) + ".acb"))
        return false;

    if (!xv2fs->SaveFile(dynamic_cast<CriwareAudioContainer *>(game_css_acb_en), std::string(GAME_CSS_SOUND_EN_PATH) + ".acb"))
        return false;

    if (!xv2fs->SaveFile(game_css_awb_jp, std::string(GAME_CSS_SOUND_JP_PATH) + ".awb"))
        return false;

    if (!xv2fs->SaveFile(game_css_awb_en, std::string(GAME_CSS_SOUND_EN_PATH) + ".awb"))
        return false;

    return true;
}

bool Xenoverse2::GetCharaCodeFromId(uint32_t id, std::string &code)
{
    if (!game_cms)
        return false;

    for (CmsEntry *entry : *game_cms)
    {
        CmsEntryXV2 *xv2_entry = dynamic_cast<CmsEntryXV2 *>(entry);
        if (!xv2_entry)
            continue;

        if (xv2_entry->id == id)
        {
            code = xv2_entry->name;
            return true;
        }
    }

    return false;
}

bool Xenoverse2::IsOriginalChara(const std::string &short_name)
{
    if (original_chars.size() != 114)
        DPRINTF("Correct the list!");

    std::string upper_name = Utils::ToUpperCase(short_name);
    return (std::find(original_chars.begin(), original_chars.end(), upper_name) != original_chars.end());
}

static bool InjectSwfAndDelete(const std::string &path)
{
    SwfFile swf;

    bool ret = swf.LoadFromFile(path);
    Utils::RemoveFile(path);

    if (!ret)
        return false;

    uint8_t *abc = nullptr;
    uint32_t abc_size = 0;

    for (uint32_t i = 0; i < swf.GetNumBlocks(); i++)
    {
        SwfDoABC *abc_block = dynamic_cast<SwfDoABC *>(swf.GetBlock(i));

        if (abc_block)
        {
            abc_size = (uint32_t)abc_block->abc.size();
            abc = new uint8_t[abc_size];
            memcpy(abc, abc_block->abc.data(), abc_size);
            break;
        }
    }

    if (!abc)
    {
        DPRINTF("That swf file doesn't have ActionSript 3 data.\n");
        return false;
    }

    ret = charasele->SetAbcBlob(abc, abc_size);
    delete[] abc;
    return ret;
}

bool Xenoverse2::CompileCharaSel(const std::string &compiler)
{
    std::vector<std::string> args;

    if (!chara_list || !charasele)
        return false;

    if (!chara_list->SaveToFile(chasel_path + CHARALIST_PATH))
        return false;

    args.push_back("-compiler.source-path=" + Utils::WindowsPath(chasel_path));
    args.push_back("-omit-trace-statements=false");
    args.push_back("-compress=false");
    args.push_back(Utils::WindowsPath(chasel_path + TIMELINE_PATH));

    int ret = Utils::RunProgram(Utils::WindowsPath(compiler), args);
    if (ret != 0)
    {
        DPRINTF("%s: Compiler failed. Ret=%d\n"
                "A ret of -1 may indicate a non existing compiler.\n"
                "A ret of 1 may indicate an error in the compile process.\n"
                "A ret of 2 may indicate that 32 bits version of Java is not installed.\n"
                "A ret of 6 may indicate problems between the compiler and the java installation.\n"
                "Compiler path=%s\n"
                "chasel_path (relative) = %s\n\n"
                "See the troubleshooting guide at installer thread.\n", FUNCNAME, ret, Utils::WindowsPath(compiler).c_str(), Utils::WindowsPath(chasel_path).c_str());
        return false;
    }

    // We don't need this anymore
    Utils::RemoveFile(chasel_path+CHARALIST_PATH);

    const std::string comp_result = chasel_path + COMPRESULT_PATH;

    if (!Utils::FileExists(comp_result))
    {
        DPRINTF("%s: swf doesn't exist after compile.\n", FUNCNAME);
        return false;
    }

    // Better do this in a separate function, so that installer can undo things easily if something fails
    /*if (!chara_list->SaveToFile(chalist_path, true, true))
        return false;*/

    return InjectSwfAndDelete(comp_result);
}

bool Xenoverse2::GetCharaName(const std::string &code, std::string &name, int lang, int index)
{
    if (!xv2fs || game_chara_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "chara_" + code + "_" + Utils::ToStringAndPad(index, 3);
    MsgEntry *entry = game_chara_names[lang]->FindEntryByName(entry_name);

    if (!entry && index != 0)
    {
        entry_name = "chara_" + code + "_000";
        entry = game_chara_names[lang]->FindEntryByName(entry_name);
    }

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

bool Xenoverse2::SetCharaName(const std::string &code, const std::string &name, int lang, int index)
{
    if (!xv2fs || game_chara_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "chara_" + code + Utils::ToStringAndPad(index, 3);
    MsgEntry *entry = game_chara_names[lang]->FindEntryByName(entry_name);

    if (entry)
    {
        entry->lines.resize(1);
        entry->lines[0] = EscapeHtml(name);
    }
    else
    {
        MsgEntry entry;

        entry.name = entry_name;
        entry.lines.push_back(EscapeHtml(name));

        if (!game_chara_names[lang]->AddEntry(entry, true))
            return false;
    }

    return true;
}

bool Xenoverse2::RemoveCharaName(const std::string &code, int lang, int index)
{
    if (!xv2fs || game_chara_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "chara_" + code + Utils::ToStringAndPad(index, 3);
    return game_chara_names[lang]->RemoveEntry(entry_name);
}

bool Xenoverse2::RemoveAllCharaName(const std::string &code, int lang)
{
    for (int i = 0; i < 1000; i++)
    {
        if (!RemoveCharaName(code, lang, i))
            return false;
    }

    return true;
}

bool Xenoverse2::GetCharaCostumeName(const std::string &code, int var, int model_preset, std::string &name, int lang)
{
    if (!xv2fs || game_chara_costume_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "variation_" + code + "_" + Utils::ToStringAndPad(var, 3) + "_" + Utils::ToStringAndPad(model_preset, 3);
    MsgEntry *entry = game_chara_costume_names[lang]->FindEntryByName(entry_name);

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

bool Xenoverse2::SetCharaCostumeName(const std::string &code, int var, int model_preset, const std::string &name, int lang)
{
    if (!xv2fs || game_chara_costume_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "variation_" + code + "_" + Utils::ToStringAndPad(var, 3) + "_" + Utils::ToStringAndPad(model_preset, 3);
    MsgEntry *entry = game_chara_costume_names[lang]->FindEntryByName(entry_name);

    if (entry)
    {
        entry->lines.resize(1);
        entry->lines[0] = EscapeHtml(name);
    }
    else
    {
        MsgEntry entry;

        entry.name = entry_name;
        entry.lines.push_back(EscapeHtml(name));

        if (!game_chara_costume_names[lang]->AddEntry(entry, true))
            return false;
    }

    return true;
}

bool Xenoverse2::RemoveCharaCostumeName(const std::string &code, int var, int model_preset, int lang)
{
    if (!xv2fs || game_chara_costume_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "variation_" + code + "_" + Utils::ToStringAndPad(var, 3) + "_" + Utils::ToStringAndPad(model_preset, 3);
    return game_chara_costume_names[lang]->RemoveEntry(entry_name);
}

bool Xenoverse2::GetSuperSkillName(uint16_t name_id, std::string &name, int lang)
{
    if (!xv2fs || sup_sk_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "spe_skill_" + Utils::ToStringAndPad(name_id, 4);
    MsgEntry *entry = sup_sk_names[lang]->FindEntryByName(entry_name);

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

bool Xenoverse2::GetUltimateSkillName(uint16_t name_id, std::string &name, int lang)
{
    if (!xv2fs || ult_sk_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "ult_" + Utils::ToStringAndPad(name_id, 4);
    MsgEntry *entry = ult_sk_names[lang]->FindEntryByName(entry_name);

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

bool Xenoverse2::GetEvasiveSkillName(uint16_t name_id, std::string &name, int lang)
{
    if (!xv2fs || eva_sk_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "avoid_skill_" + Utils::ToStringAndPad(name_id, 4);
    MsgEntry *entry = eva_sk_names[lang]->FindEntryByName(entry_name);

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

bool Xenoverse2::GetAwakenSkillName(uint16_t name_id, std::string &name, int lang)
{
    if (!xv2fs || awa_sk_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "met_skill_" + Utils::ToStringAndPad(name_id, 4);
    MsgEntry *entry = awa_sk_names[lang]->FindEntryByName(entry_name);

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

uint8_t *Xenoverse2::GetSelPortrait(const std::string &name, size_t *psize)
{
    if (!game_sel_port)
        return nullptr;

    uint16_t idx = game_sel_port->FindIndexByName(name);
    if (idx == 0xFFFF)
        return nullptr;

    const EmbContainedFile &file = (*game_sel_port)[idx];

    *psize = file.GetSize();
    uint8_t *buf = new uint8_t[*psize];

    memcpy(buf, file.GetData(), *psize);
    return buf;
}

bool Xenoverse2::SetSelPortrait(const std::string &name, const uint8_t *buf, size_t size)
{
    if (!game_sel_port)
        return false;

    uint16_t idx = game_sel_port->FindIndexByName(name);
    if (idx != 0xFFFF)
        return game_sel_port->ReplaceFile(idx, buf, size);

    game_sel_port->AppendFile(buf, size, name);
    return true;
}

bool Xenoverse2::RemoveSelPortrait(const std::string &name)
{
    if (!game_sel_port)
        return false;

    uint16_t idx = game_sel_port->FindIndexByName(name);
    if (idx == 0xFFFF)
        return true;

    game_sel_port->RemoveFile(idx);
    return true;
}

bool Xenoverse2::GetLobbyName(uint32_t name_id, std::string &name, int lang)
{
    if (!xv2fs || game_lobby_texts.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "name_" + Utils::ToStringAndPad(name_id, 4);
    MsgEntry *entry = game_lobby_texts[lang]->FindEntryByName(entry_name);

    if (!entry)
        return false;

    name = UnescapeHtml(entry->lines[0]);
    return true;
}

bool Xenoverse2::SetLobbyName(uint32_t name_id, const std::string &name, int lang)
{
    if (!xv2fs || game_lobby_texts.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "name_" + Utils::ToStringAndPad(name_id, 4);
    MsgEntry *entry = game_lobby_texts[lang]->FindEntryByName(entry_name);

    if (entry)
    {
        entry->lines.resize(1);
        entry->lines[0] = EscapeHtml(name);
    }
    else
    {
        MsgEntry entry;

        entry.name = entry_name;
        entry.lines.push_back(EscapeHtml(name));

        if (!game_lobby_texts[lang]->AddEntry(entry, true))
            return false;
    }

    return true;
}

bool Xenoverse2::RemoveLobbyName(uint32_t name_id, int lang)
{
    if (!xv2fs || game_lobby_texts.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    std::string entry_name = "name_" + Utils::ToStringAndPad(name_id, 4);
    return game_lobby_texts[lang]->RemoveEntry(entry_name);
}

bool Xenoverse2::GetCacCostumeName(uint32_t name_idx, std::string &name, int lang)
{
    if (!xv2fs || game_cac_costume_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    if (name_idx >= game_cac_costume_names[lang]->GetNumEntries())
        return false;

    const MsgEntry &entry = (*game_cac_costume_names[lang])[name_idx];

    name = UnescapeHtml(entry.lines[0]);
    return true;
}

bool Xenoverse2::GetAccesoryName(uint32_t name_idx, std::string &name, int lang)
{
    if (!xv2fs || game_accesory_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    if (name_idx >= game_accesory_names[lang]->GetNumEntries())
        return false;

    const MsgEntry &entry = (*game_accesory_names[lang])[name_idx];

    name = UnescapeHtml(entry.lines[0]);
    return true;
}

bool Xenoverse2::GetTalismanName(uint32_t name_idx, std::string &name, int lang)
{
    if (!xv2fs || game_talisman_names.size() != XV2_LANG_NUM || lang < 0 || lang >= XV2_LANG_NUM)
        return false;

    if (name_idx >= game_talisman_names[lang]->GetNumEntries())
        return false;

    const MsgEntry &entry = (*game_talisman_names[lang])[name_idx];

    name = UnescapeHtml(entry.lines[0]);
    return true;
}

HcaFile *Xenoverse2::GetCssSound(uint32_t cue_id, bool english)
{
    AcbFile *acb = (english) ? game_css_acb_en : game_css_acb_jp;
    Afs2File *awb = (english) ? game_css_awb_en : game_css_awb_jp;

    if (!acb || !awb)
        return nullptr;

    uint32_t cue_idx = acb->CueIdToCueIndex(cue_id);
    if (cue_idx == (uint32_t)-1)
        return nullptr;

    bool external;
    uint32_t awb_idx = acb->CueIndexToAwbIndex(cue_idx, &external);    

    if (awb_idx == (uint32_t)-1 || !external)
        return nullptr;

    uint64_t size;
    uint8_t *hca_buf = awb->ExtractFile(awb_idx, &size);

    if (hca_buf)
    {
        HcaFile *hca = new HcaFile();

        bool ret = hca->Load(hca_buf, size);
        delete[] hca_buf;

        if (ret)
            return hca;

        delete hca;
    }

    return nullptr;
}

HcaFile *Xenoverse2::GetCssSound(const std::string &name, bool english)
{
    AcbFile *acb = (english) ? game_css_acb_en : game_css_acb_jp;
    Afs2File *awb = (english) ? game_css_awb_en : game_css_awb_jp;

    if (!acb || !awb)
        return nullptr;

    uint32_t cue_id = acb->FindCueId(name);
    if (cue_id == (uint32_t)-1)
        return nullptr;

    return GetCssSound(cue_id, english);
}

bool Xenoverse2::SetCssSound(uint32_t cue_id, bool english, HcaFile &hca)
{
    AcbFile *acb = (english) ? game_css_acb_en : game_css_acb_jp;
    Afs2File *awb = (english) ? game_css_awb_en : game_css_awb_jp;

    if (!acb || !awb)
        return nullptr;

    uint32_t cue_idx = acb->CueIdToCueIndex(cue_id);
    if (cue_idx == (uint32_t)-1)
        return nullptr;

    uint32_t track_idx = acb->CueIndexToTrackIndex(cue_idx);
    if (track_idx == (uint32_t)-1)
        return nullptr;

    bool external;
    uint32_t awb_idx = acb->CueIndexToAwbIndex(cue_idx, &external);

    if (awb_idx == (uint32_t)-1 || !external)
        return nullptr;    

    if (!acb->SetCueLength(cue_id, (uint32_t)(hca.GetDuration()*1000.0f)))
        return false;

    if (!acb->SetWaveformNumSamples(track_idx, hca.GetNumSamples()))
        return false;

    size_t size;
    uint8_t *buf = hca.Save(&size);
    if (!buf)
        return nullptr;

    if (!awb->SetFile(awb_idx, buf, size, true))
    {
        delete[] buf;
        return false;
    }

    return true;
}

uint32_t Xenoverse2::SetCssSound(const std::string &name, bool english, HcaFile &hca)
{
    AcbFile *acb = (english) ? game_css_acb_en : game_css_acb_jp;
    Afs2File *awb = (english) ? game_css_awb_en : game_css_awb_jp;

    if (!acb || !awb)
        return (uint32_t)-1;

    uint32_t cue_id = acb->FindCueId(name);
    if (cue_id != (uint32_t)-1)
    {
        if (!SetCssSound(cue_id, english, hca))
            return (uint32_t)-1;

        return cue_id;
    }

    cue_id = acb->FindCueId("X2_FREE");
    if (cue_id != (uint32_t)-1)
    {
        if (!SetCssSound(cue_id, english, hca))
            return (uint32_t)-1;

        if (!acb->SetCueName(cue_id, name))
            return (uint32_t)-1;

        return cue_id;
    }

    uint16_t awb_idx = (uint16_t)awb->GetNumFiles();

    if (!acb->AddTrack(name, awb_idx, true, (uint32_t)(hca.GetDuration()*1000.0f), hca.GetNumSamples(),
                       false, hca.GetSampleRate(), (uint8_t)hca.GetNumChannels(), 2, 6, &cue_id, nullptr, nullptr))
    {
        return (uint32_t)-1;
    }

    size_t size;
    uint8_t *buf = hca.Save(&size);
    if (!buf)
        return (uint32_t)-1;

    if (!awb->AddFile(buf, size, true))
    {
        delete[] buf;
        return (uint32_t)-1;
    }

    return cue_id;
}

bool Xenoverse2::FreeCssSound(const std::string &name, bool english)
{
    AcbFile *acb = (english) ? game_css_acb_en : game_css_acb_jp;
    if (!acb)
        return false;

    uint32_t cue_id = acb->FindCueId(name);
    if (cue_id == (uint32_t)-1)
        return true; // Yes, true

    if (!acb->SetCueName(cue_id, "X2_FREE"))
        return false;

    return true;
}

std::string Xenoverse2::GetCharaAndCostumeName(const std::string &code, uint32_t costume_id, uint32_t model_preset, int lang)
{
    std::string result, name, costume;

    if (!Xenoverse2::GetCharaName(code, name, lang, costume_id))
        name = code;

    if (costume_id == 0xFFFFFFFF)
        costume = "[Wildcard]";

    else if (!Xenoverse2::GetCharaCostumeName(code, costume_id, model_preset, costume, lang))
        costume = Utils::UnsignedToString(costume_id, true);

    result = name + " - " + costume;
    return result;
}

static std::string dummy_costume(uint32_t costume_id)
{
    if (costume_id == 0xFFFFFFFF)
        return "[Wildcard]";

    return Utils::UnsignedToString(costume_id, true);
}

std::string Xenoverse2::GetCharaAndCostumeName(uint32_t char_id, uint32_t costume_id, uint32_t model_preset, bool ignore_unknown, int lang)
{
    CmsEntry *cms_entry;

    if (char_id == 0xFFFFFFFF)
    {
        return "[Wildcard] - " + dummy_costume(costume_id);
    }

    if (!game_cms || ((cms_entry = game_cms->FindEntryByID(char_id)) == nullptr))
    {
        if (ignore_unknown)
            return std::string();

        return Utils::UnsignedToString(char_id, true) + " - " + dummy_costume(costume_id);
    }

    return GetCharaAndCostumeName(cms_entry->name, costume_id, model_preset, lang);
}

std::string Xenoverse2::GetCacTopName(uint16_t item_id, int lang)
{
    if (item_id == 0xFFFF)
        return "None";

    if (!game_top_idb)
        return std::string();

    IdbEntry *entry = game_top_idb->FindEntryByID(item_id);
    if (!entry)
        return std::string();

    std::string ret;
    GetCacCostumeName(entry->name_id, ret, lang);

    return ret;
}

std::string Xenoverse2::GetCacBottomName(uint16_t item_id, int lang)
{
    if (item_id == 0xFFFF)
        return "None";

    if (!game_bottom_idb)
        return std::string();

    IdbEntry *entry = game_bottom_idb->FindEntryByID(item_id);
    if (!entry)
        return std::string();

    std::string ret;
    GetCacCostumeName(entry->name_id, ret, lang);

    return ret;
}

std::string Xenoverse2::GetCacGlovesName(uint16_t item_id, int lang)
{
    if (item_id == 0xFFFF)
        return "None";

    if (!game_gloves_idb)
        return std::string();

    IdbEntry *entry = game_gloves_idb->FindEntryByID(item_id);
    if (!entry)
        return std::string();

    std::string ret;
    GetCacCostumeName(entry->name_id, ret, lang);

    return ret;
}

std::string Xenoverse2::GetCacShoesName(uint16_t item_id, int lang)
{
    if (item_id == 0xFFFF)
        return "None";

    if (!game_shoes_idb)
        return std::string();

    IdbEntry *entry = game_shoes_idb->FindEntryByID(item_id);
    if (!entry)
        return std::string();

    std::string ret;
    GetCacCostumeName(entry->name_id, ret, lang);

    return ret;
}

std::string Xenoverse2::GetCacAccesoryName(uint16_t item_id, int lang)
{
    if (item_id == 0xFFFF)
        return "None";

    if (!game_accesory_idb)
        return std::string();

    IdbEntry *entry = game_accesory_idb->FindEntryByID(item_id);
    if (!entry)
        return std::string();

    std::string ret;
    GetAccesoryName(entry->name_id, ret, lang);

    return ret;
}

std::string Xenoverse2::GetTalismanNameEx(uint16_t item_id, int lang)
{
    if (item_id == 0xFFFF)
        return "None";

    if (!game_talisman_idb)
        return std::string();

    IdbEntry *entry = game_talisman_idb->FindEntryByID(item_id);
    if (!entry)
        return std::string();

    std::string ret;
    GetTalismanName(entry->name_id, ret, lang);

    return ret;
}





